from time import time
from itertools import combinations
from math import inf, floor
from ..interface import filer, messenger
from ..models.interpreters import DAT_Interpreter
from ..models.ansi_style import Color, Effect

# Funções de Grafos
def get_level_structure(node: int, graph: dict) -> list[dict]:
   # Variáveis Iniciais
   level_structure = list()
   max_level_length = 0
   leveled_nodes = {node}

   # Calculando Nível Inicial
   node_degree = len(graph[node])
   current_level = {node: node_degree}

   # Loop para cada Nível
   while 1:
      # Adicionando Nível Atual à Lista de Níveis
      level_structure.append(current_level)
      next_level = dict()

      # Verificando Adjacência do Nível Atual
      for n in current_level:
         adjacent_nodes = graph[n]
         for a in adjacent_nodes:
            if a not in leveled_nodes:
               leveled_nodes.add(a)
               next_level[a] = len(graph[a])
      
      # Verificando se o Nível Atual é Vazio
      if len(next_level) == 0:
         break

      # Verificando se o Comprimento do Nível é o Maior
      if len(next_level) > max_level_length:
         max_level_length = len(next_level)
      
      # Intercambiando Níveis
      current_level = next_level
   
   return level_structure, max_level_length

def get_pseudo_peripheral_node_pair(graph: dict) -> list[int]:
   # Encontrando Node de Menor Grau como Chute Inicial
   n_nodes = len(graph)
   pseudo_peripheral_node_1 = -1
   pseudo_peripheral_node_2 = -1
   min_degree = inf
   for node in range(n_nodes):
      degree = len(graph[node])
      if degree < min_degree:
         min_degree = degree
         pseudo_peripheral_node_1 = node

   # Construindo Estrutura de Nível do Node Chutado
   level_structure, _ = get_level_structure(pseudo_peripheral_node_1, graph)
   
   # Processo de Seleção (Só termina quando Tivermos um Node Inicial e Outro Final)
   while pseudo_peripheral_node_2 == -1:
      # Inicializando um Comprimento Mínimo para o Nível de Maior Comprimento
      min_level_length = inf

      # Ordenando Último Nível pelo Grau
      ordered_level = level_structure[-1]
      ordered_level = sorted(ordered_level.keys(), key = lambda k: ordered_level[k])

      # Avaliando Canditados
      level_structure_length = len(level_structure)
      n_candidates = floor(len(ordered_level) / 2) + 1
      for candidate_node in ordered_level[:n_candidates]:
         # Construindo Estrutura de Nível do Node Candidato
         candidate_level_structure, candidate_max_level_length = get_level_structure(candidate_node, graph)

         # Verificando se Canditado pode Substituir o Node Escolhido Atual
         if candidate_max_level_length < min_level_length:
               # Avaliando se o Node pode ser Inicial ou Final
               if len(candidate_level_structure) > level_structure_length:
                  level_structure = candidate_level_structure
                  pseudo_peripheral_node_1 = candidate_node
                  break
               else:
                  pseudo_peripheral_node_2 = candidate_node
                  min_level_length = candidate_max_level_length

   return pseudo_peripheral_node_1, pseudo_peripheral_node_2

def get_bandwidth(graph: dict) -> list[int]:
   bandwidth = [0] * len(graph)
   for node, adjacent_nodes in graph.items():
      min_node = min(adjacent_nodes)
      bandwidth[node] = node - min_node if min_node < node else 0
   return bandwidth

# Funções de Reordenação 
def reorder_rcm(graph: dict) -> list[int]:
   # Variáveis Iniciais
   n_nodes = len(graph)
   new_order = [0] * n_nodes

   # Encontrando Node de Menor Grau como Node Pseudo-Periférico
   pseudo_peripheral_node = 0
   min_degree = len(graph[0])
   for node in range(1, n_nodes):
      degree = len(graph[node])
      if degree < min_degree:
         min_degree = degree
         pseudo_peripheral_node = node

   # Construindo Estrutura de Nível do Node Pseudo-Periférico
   level_structure, _ = get_level_structure(pseudo_peripheral_node, graph)
   
   # Percorrendo Níveis
   order = n_nodes - 1
   for level in level_structure:
      # Ordenando Nodes do Nível por Grau
      level = sorted(level.keys(), key = lambda k: level[k])

      # Registrando Nova Ordem
      for node in level:
         new_order[node] = order
         order -= 1
   
   return new_order

def reorder_sloan(graph: dict) -> list[int]:
   # Encontrando Nodes Pseudo-Periféricos Inicial e Final
   node_start, node_end = get_pseudo_peripheral_node_pair(graph)
   n_nodes = len(graph)
   new_order = [0] * n_nodes

   # Calculando Lista de Distância para Node Final
   level_structure_end, _ = get_level_structure(node_end, graph)
   distances_to_end = [0] * n_nodes
   degrees = [0] * n_nodes
   for level_order, level in enumerate(level_structure_end):
      for node in level.keys():
         distances_to_end[node] = level_order
         degrees[node] = level[node]
   del level_structure_end

   # Atribuindo Status aos Nodes
   # 0 - posactive
   # 1 - active
   # 2 - preactive
   # 3 - inactive
   status = [3] * n_nodes

   # Atribuindo Prioridades
   W1 = 2
   W2 = 1
   priorities = list()
   for i in range(n_nodes):
      current_degree = degrees[i] + 1
      p = (n_nodes - current_degree) * W1 + distances_to_end[i] * W2
      priorities.append(p)
   
   # Iniciando Pilha de Nodes Elegíveis
   eligible_nodes = [node_start]
   status[node_start] = 2
   
   # Processo de Reordenação
   order = 0
   while len(eligible_nodes) != 0:
      # Ordenando Nodes por Prioridade
      eligible_nodes = sorted(eligible_nodes, key = lambda k: priorities[k])

      # Selecionando Node de Maior Prioridade
      highest_priority_node = eligible_nodes.pop()

      # Verificando se Node de Maior Prioridade é Pré-ativo
      if status[highest_priority_node] == 2:
         # Examinando Nodes Adjacentes
         adjacent_nodes_level_1 = graph[highest_priority_node]
         for a1 in adjacent_nodes_level_1:
               # Alterando Prioridade
               priorities[a1] += W1

               # Adicionando Node à Lista de Nodes Elegíveis e Atualizando Status se é Inativo
               if status[a1] == 3:
                  status[a1] = 2
                  eligible_nodes.append(a1)

      # Adicionando Nova Ordem ao Node de Maior Prioridade
      status[highest_priority_node] = 0
      new_order[highest_priority_node] = order
      order += 1

      # Examinando Nodes Adjacentes
      adjacent_nodes_level_1 = graph[highest_priority_node]
      for a1 in adjacent_nodes_level_1:
         # Verificando se Node de Maior Prioridade é Pré-ativo
         if status[a1] == 2:
               # Alterando Prioridade e Status
               priorities[a1] += W1
               status[a1] = 1

               # Examinando Nodes Adjacentes
               adjacent_nodes_level_2 = graph[a1]
               for a2 in adjacent_nodes_level_2:
                  # Verificando se Node de Maior Prioridade é Ativo ou Pré-ativo
                  if (status[a2] == 1) or (status[a2] == 2):
                     # Alterando Prioridade
                     priorities[a2] += W1

                  # Verificando se Node de Maior Prioridade é Inativo
                  elif status[a2] == 3:
                     # Alterando Prioridade, Status e Adicionando à Lista de Nodes Elegíveis
                     priorities[a2] += W1
                     status[a2] = 2
                     eligible_nodes.append(a2)

   return new_order

def reorder_boost_rcm(graph: dict) -> list[int]:
   # Tentando Importar Módulo
   try:
      from ..resources.shared.boost import reorder_rcm as r_rcm
   except ModuleNotFoundError:
      raise OSError('Your Operational System has no support for this Reordering Method.')
   
   # Gerando Reordenação
   new_order = [0] * len(graph)
   r_rcm(graph, new_order)
   
   return new_order

def reorder_boost_sloan(graph: dict) -> list[int]:
   # Tentando Importar Módulo
   try:
      from ..resources.shared.boost import reorder_sloan as r_sloan
   except ModuleNotFoundError:
      raise OSError('Your Operational System has no support for this Reordering Method.')
   
   # Gerando Reordenação
   new_order = [0] * len(graph)
   r_sloan(graph, new_order)
   
   return new_order

# Métodos de Reordenação Suportados
supported_methods = {
   'rcm': reorder_rcm,
   'sloan': reorder_sloan,
   'boost_rcm': reorder_boost_rcm,
   'boost_sloan': reorder_boost_sloan
}

# Função de Inicialização
def start(method: str, dat_path: str, flags: dict[str, str]):
   # Verificando se o Método de Reordenação é Suportado
   try:
      reordering_function = supported_methods[method]
   except KeyError:
      raise KeyError(f'The method "{method}" is not supported.')

   # Lendo Arquivo .dat
   time_read = time()
   dat_data = filer.read(dat_path)

   # Interpretando Informações
   dat_interpreter = DAT_Interpreter()
   dat_interpreter.read_nodes(dat_data)
   dat_interpreter.read_patches(dat_data)
   dat_interpreter.read_elements(dat_data)
   model = dat_interpreter.model

   # Criando Matriz do Grafo para Reordenação
   n = len(model.nodes)
   graph = {i: set() for i in range(n)}
   for group in model.element_groups.values():
      for element in group.elements.values():
         for i, j in combinations(element.node_ides, 2):
            i -= 1
            j -= 1
            graph[i].add(j)
            graph[j].add(i)
   time_read = time() - time_read

   # Reordenando
   time_reorder = time()
   new_order = reordering_function(graph)
   time_reorder = time() - time_reorder

   # Adicionando Reordenação ao Modelo de Simulação
   time_write = time()
   new_order = [n + 1 for n in new_order]
   dat_interpreter.model.node_solver_order = new_order

   # Gerando e Incluindo Codificação da Ordem no Arquivo .dat
   order_data = dat_interpreter.write_node_solver_order()
   dat_data = dat_data.replace('%ELEMENT\n', order_data[1:] + '\n%ELEMENT\n')
   filer.write(dat_path, dat_data)
   time_write = time() - time_write

   # Verificando Flags
   fk = flags.keys()
   if ('-i' in fk) or ('--info' in fk):
      # Calculando Informações Adiconais sobre a Reordenação
      # Gerando Novo Grafo com a Reordenação
      new_graph = {i: set() for i in range(n)}
      for node, adjacent_nodes in graph.items():
         new_graph[new_order[node] - 1] = {new_order[a] - 1 for a in adjacent_nodes}

      # Calculando Larguras de Banda
      old_bandwidth = get_bandwidth(graph)
      new_bandwidth = get_bandwidth(new_graph)

      # Calculando PercentuaL de Redução de Largura de Banda Máxima
      old_max_bandwidth = max(old_bandwidth)
      new_max_bandwidth = max(new_bandwidth)
      max_bandwidth_reduction = (old_max_bandwidth - new_max_bandwidth) / old_max_bandwidth
      result_color = Color.green if max_bandwidth_reduction > 0 else Color.red
      messenger.info(f'Maximum Bandwidth Reduction = {result_color}{max_bandwidth_reduction:.2%}{Effect.reset}')

      # Calculando PercentuaL de Redução de Largura de Banda Média
      old_avg_bandwidth = sum(old_bandwidth) / n
      new_avg_bandwidth = sum(new_bandwidth) / n
      avg_bandwidth_reduction = (old_avg_bandwidth - new_avg_bandwidth) / old_avg_bandwidth
      result_color = Color.green if avg_bandwidth_reduction > 0 else Color.red
      messenger.info(f'Average Bandwidth Reduction = {result_color}{avg_bandwidth_reduction:.2%}{Effect.reset}')

      # Exibindo Tempos
      messenger.info(f'Time to Read    = {time_read:.3f} s')
      messenger.info(f'Time to Reorder = {time_reorder:.3f} s')
      messenger.info(f'Time to Write   = {time_write:.3f} s')
      messenger.info(f'Total Time      = {(time_read + time_reorder + time_write):.3f} s')


# Importações - Bibliotecas Nativas
from sys import exit

# Importações - Bibliotecas Locais
from ..models.ansi_style import Color, Effect

# Funções Globais
def show(message: str):
   print(message)

def error(message: str, name: str = 'CommandError'):
   # Exibindo Mensagem de Erro
   print(f'{Color.red}| {name} | :{Effect.reset} {message}')

   # Terminando Processo (Se não Estiver no Modo Interativo)
   from .core import in_interactive_mode
   if not in_interactive_mode:
      exit(1)

def warning(message: str):
   # Exibindo Mensagem de Alerta
   print(f'{Color.yellow}| Warning | :{Effect.reset} {message}')

def info(message: str):
   # Exibindo Mensagem de Informativa de Destaque
   print(f'{Color.blue}| Info | :{Effect.reset} {message}')


from ..interface import filer, messenger
from ..models.extraction_components import (
   Condition,
   Attribute
)

# Funções Globais
def change(attribute_name: str, attribute_value: str, dat_data: str, condition: str = None):
   # Instanciando Tabela de Resultado e Condição de Atributos
   attribute = Attribute(attribute_name)
      
   # Alterando Valor do Atributo no .dat
   attribute_condition = Condition(condition)
   changed_data = attribute.change_to(attribute_value, dat_data, attribute_condition)
   
   return changed_data

def start(changes_terms: list[str], dat_path: str, condition: str):
   # Lendo Arquivo .dat
   dat_data = filer.read(dat_path)

   # Gerando Dados com Alterações
   changed_data = change(changes_terms[0], changes_terms[1], dat_data, condition)

   # Escrevendo Alterações no .dat
   filer.write(dat_path, changed_data)


import unittest
import os

class DefaultTest(unittest.TestCase):
   def default_test(self, artifact_name: str, artifact_extension: str, test_id: str, args: list):
      # Definindo paths
      path = f'tests/benchmark/generate/{artifact_name}/'
      artifact_path = f'{path}{artifact_name}.{artifact_extension}'
      exp_path = f'{artifact_path[:-4]}_exp_{test_id}.{artifact_extension}'

      # Gerando Artefato
      args_joined = ' '.join(args)
      command = f'python -m lmcv_tools generate {artifact_name} {args_joined} {artifact_path}'
      code = os.system(command)
      self.assertEqual(code, 0, 'A geração falhou.')

      # Comparando Artefato com o Resultado Esperado
      artifact_file = open(artifact_path, 'r')
      exp_file = open(exp_path, 'r')
      artifact_data = artifact_file.read()
      exp_data = exp_file.read()
      artifact_file.close()
      exp_file.close()
      self.assertEqual(artifact_data, exp_data, 'O Artefato está incorreto.')

      # Removendo Arquivo .csv Gerado
      os.remove(artifact_path)

class TestVirtualLaminas(DefaultTest):
   def default_test(self, test_id: str, args: list):
      super().default_test('virtual_laminas', 'inp', test_id, args)

   def test_shell_element(self):
      test_id = 'shell'
      args = ['2', 'Shell', '0.5', '3', '1.0', 'voigt', '90.0', '380.0', '0.27', '0.30', '2000', '1000', 'False']
      self.default_test(test_id, args)

   def test_solid_element(self):
      test_id = 'solid'
      args = ['2', 'Solid', '0.5', '3', '1.0', 'voigt', '90.0', '380.0', '0.27', '0.30', '2000', '1000', 'False']
      self.default_test(test_id, args)

   def test_voigt_model(self):
      test_id = 'voigt'
      args = ['40', 'Solid', '0.25', '3', '1.0', 'voigt', '90.0', '380.0', '0.27', '0.30', '2000', '1000', 'False']
      self.default_test(test_id, args)
   
   def test_mori_tanaka_model(self):
      test_id = 'mori_tanaka'
      args = ['40', 'Solid', '0.25', '3', '1.0', 'mori_tanaka', '90.0', '380.0', '0.27', '0.30', '2000', '1000', 'False']
      self.default_test(test_id, args)
   
   def test_hashin_shtrikman_upper_bound_model_1(self):
      test_id = 'hs_ub_1'
      args = ['40', 'Solid', '0.25', '3', '1.0', 'hashin_shtrikman_upper_bound', '90.0', '380.0', '0.27', '0.30', '2000', '1000', 'False']
      self.default_test(test_id, args)
   
   def test_hashin_shtrikman_lower_bound_model_1(self):
      test_id = 'hs_lb_1'
      args = ['40', 'Solid', '0.25', '3', '1.0', 'hashin_shtrikman_lower_bound', '90.0', '380.0', '0.27', '0.30', '2000', '1000', 'False']
      self.default_test(test_id, args)
   
   def test_hashin_shtrikman_upper_bound_model_2(self):
      test_id = 'hs_ub_2'
      args = ['40', 'Solid', '0.25', '3', '1.0', 'hashin_shtrikman_upper_bound', '380.0', '90.0', '0.30', '0.27', '1000', '2000', 'False']
      self.default_test(test_id, args)
   
   def test_hashin_shtrikman_lower_bound_model_2(self):
      test_id = 'hs_lb_2'
      args = ['40', 'Solid', '0.25', '3', '1.0', 'hashin_shtrikman_lower_bound', '380.0', '90.0', '0.30', '0.27', '1000', '2000', 'False']
      self.default_test(test_id, args)

   def test_smart_laminas_1(self):
      test_id = 'smart_1'
      args = ['100', 'Shell', '3.5', '3', '0.2', 'voigt', '90.0', '380.0', '0.27', '0.30', '2000', '1000', 'True']
      self.default_test(test_id, args)
   
   def test_smart_laminas_2(self):
      test_id = 'smart_2'
      args = ['100', 'Shell', '3.5', '3', '5', 'voigt', '90.0', '380.0', '0.27', '0.30', '2000', '1000', 'True']
      self.default_test(test_id, args)

class TestRectangle(DefaultTest):
   def default_test(self, test_id: str, args: list):
      super().default_test('rectangle', 'dat', test_id, args)

   def test_q8(self):
      test_id = 'q8'
      args = ['Q8', '16', '9', '8', '4']
      self.default_test(test_id, args)

class TestCuboid(DefaultTest):
   def default_test(self, test_id: str, args: list):
      super().default_test('cuboid', 'dat', test_id, args)

   def test_brick20_cube(self):
      test_id = 'brick20_cube'
      args = ['BRICK20', '1.0', '1.0', '1.0', '1', '1', '1']
      self.default_test(test_id, args)

   def test_brick20_cuboid(self):
      test_id = 'brick20_cuboid'
      args = ['BRICK20', '0.333', '0.6789', '1.314159', '2', '3', '4']
      self.default_test(test_id, args)

class TestNURBS_Rectangle(DefaultTest):
   def default_test(self, test_id: str, args: list):
      super().default_test('nurbs_rectangle', 'dat', test_id, args)

   def test_nurbs_square(self):
      test_id = 'nurbs_square'
      args = ['3', '3', '1.0', '1.0', '1', '1']
      self.default_test(test_id, args)

   def test_nurbs_rectangle(self):
      test_id = 'nurbs_rectangle'
      args = ['2', '5', '0.333', '1.314159', '7', '9']
      self.default_test(test_id, args)

class TestNURBS_Cuboid(DefaultTest):
   def default_test(self, test_id: str, args: list):
      super().default_test('nurbs_cuboid', 'dat', test_id, args)

   def test_nurbs_cube(self):
      test_id = 'nurbs_cube'
      args = ['3', '3', '3', '1.0', '1.0', '1.0', '1', '1', '1']
      self.default_test(test_id, args)

   def test_nurbs_cuboid(self):
      test_id = 'nurbs_cuboid'
      args = ['2', '7', '5', '0.333', '0.6789', '1.314159', '2', '3', '4']
      self.default_test(test_id, args)

class TestCylindricalPanel(DefaultTest):
   def default_test(self, test_id: str, args: list):
      super().default_test('cyl_panel', 'dat', test_id, args)

   def test_semi_cyl_panel(self):
      test_id = 'semi_cyl_panel'
      args = ['Q8', '5.0', '3', '5', '1.5', '180.0', '0.0']
      self.default_test(test_id, args)

class TestSlitAnnularPlate(DefaultTest):
   def default_test(self, test_id: str, args: list):
      super().default_test('slit_annular_plate', 'dat', test_id, args)

   def test_semi_cyl_panel(self):
      test_id = 'q8_6_10_2x6'
      args = ['Q8', '6', '10', '2', '6']
      self.default_test(test_id, args)


# Importações - Bibliotecas Locais
from re import match
from . import messenger
from ..models.custom_errors import CommandError

# Constantes Globais
version = '0.0.28'
in_interactive_mode = False
message_welcome = '''
LMCV Tools is a command line tool that provides a series of useful functionali-
ties for the day-to-day simulations of the "Laboratório de Mecânica Computacio-
nal e Visualização" of the "Universidade Federal do Ceará" (UFC). 

Since a command was not typed, interactive mode was started. Here, you can type
multiple commands in sequence interactively:

To get help | type the command "help"
To exit     | type the command "exit"
'''

# Funções Pré-processamento de Comandos
def show_version(args: list[str] = None, flags: dict[str, str] = None):
   messenger.show(f'LMCV Tools - v{version}')

def pre_help(args: list[str], flags: dict[str, str] = None):   
   from ..commands import help
   
   if len(args) > 0:
      help.start(args[0])
   else:
      help.start()

def pre_translate(args: list[str], flags: dict[str, str]):
   from ..commands import translate

   # Verificando Path do Input
   try:
      input_path = args[0]
   except IndexError:
      raise CommandError('An Input File Path is required.')

   # Verificando Extensão do Output
   try:
      if 'to' != args[1]:
         raise CommandError('"to" keyword after Input File Path is required.')
   except IndexError:
      raise CommandError('"to" keyword after Input File Path is required.')
   try:
      output_extension = args[2]
   except IndexError:
      raise CommandError('An Extension for Output File after "to" keyword is required.')
   
   # Iniciando Tradução
   translate.start(input_path, output_extension, args[3:], flags)

def pre_extract(terms: list[str], flags: dict[str, str]):
   from ..commands import extract

   # Verificando Sintaxe Básica da Sentença
   if 'from' not in terms:
      raise CommandError('The keyword "from" is required.', help=True)
   
   # Verificando se ao menos 1 Atributo foi fornecido
   index = terms.index('from')
   attributes = terms[:index]
   if len(attributes) == 0:
      raise CommandError('At least one attribute before "from" is required.')

   # Verificando se o Path do Arquivo .pos foi fornecido
   try:
      pos_path = terms[index + 1]
   except IndexError:
      raise CommandError('The path to .pos file after "from" is required.')
   
   # Verificando se um Path para o .csv foi fornecido
   csv_path = pos_path[:-3] + 'csv'
   index_to = 0
   if 'to' in terms[index + 1:]:
      index_to = terms.index('to')
      try:
         csv_path = terms[index_to + 1]
      except IndexError:
         raise CommandError('The Syntax "to [path/to/.csv]" is optional, but it is incomplete.')
   
   # Verificando se uma Condição foi fornecida
   condition = None
   if 'where' in terms[index + 1:]:
      index_where = terms.index('where')
      if index_to > index_where:
         condition = terms[index_where + 1:index_to]
      else:
         condition = terms[index_where + 1:]
      if len(condition) == 0:
         raise CommandError('The Syntax "where [condition]" is optional, but it is incomplete.')
      condition = ' '.join(condition)

   # Extraindo Itens do Arquivo .pos
   extract.start(attributes, pos_path, condition, csv_path)

def pre_generate(args: list[str], flags: dict[str, str]):
   from ..commands import generate
   
   # Verificando se um Artefato foi Dado
   if len(args) == 0:
      raise CommandError('An Artifact must be given.', help=True)

   # Iniciando Comando
   artifact = args[0]
   generate.start(artifact, args[1:])

def pre_reorder(args: list[str], flags: dict[str, str]):
   from ..commands import reorder
   
   # Verificando se um Método de Reordenação foi Fornecido
   try:
      method = args[0]
   except IndexError:
      raise CommandError('A reorder method is required.')
   
   # Verificando se um Caminho de Arquivo foi Fornecido
   try:
      dat_path = args[1]
   except IndexError:
      raise CommandError('A path to .dat file is required.')
   
   # Executando Reordenação
   reorder.start(method, dat_path, flags)

def pre_change(terms: list[str], flags: dict[str, str]):
   from ..commands import change

   # Verificando Sintaxe Básica da Sentença
   if 'in' not in terms:
      raise CommandError('The keyword "in" is required.', help=True)
   
   # Verificando se a Mudança foi fornecida
   index = terms.index('in')
   changes_terms = terms[:index]
   if len(changes_terms) == 0:
      raise CommandError('One change before "in" is required.')
   
   # Verificando Sintaxe da Mudança
   changes_terms = ' '.join(changes_terms)
   changes_terms = changes_terms.split(' to ')
   if len(changes_terms) != 2:
      raise CommandError('The Syntax chosen to write a change is incorrect.')

   # Verificando se o Path do Arquivo .dat foi fornecido
   try:
      dat_path = terms[index + 1]
   except IndexError:
      raise CommandError('The path to .dat file after "in" is required.')
   
   # Verificando se uma Condição foi fornecida
   condition = None
   if 'where' in terms[index + 1:]:
      index_where = terms.index('where')
      condition = terms[index_where + 1:]
      if len(condition) == 0:
         raise CommandError('The Syntax "where [condition]" is optional, but it is incomplete.')
      condition = ' '.join(condition)

   # Alterando Atributosno Arquivo .dat
   change.start(changes_terms, dat_path, condition)

# Relação Comando/Função
commands = {
   'version': show_version,
   'help': pre_help,
   'translate': pre_translate,
   'extract': pre_extract,
   'generate': pre_generate,
   'reorder': pre_reorder,
   'change': pre_change
}

# Funções de Inicialização
def execute_command(name: str, args: list[str]):
   # Encapsulando Execução para Tratamento Padrão de Erros
   try:
      # Tentando Identificar o Comando
      try:
         command_function = commands[name]
      except KeyError:
         raise CommandError('Unknown command.', help=True)

      # Analisando a Presença de Flags
      flags = dict[str, str]()
      flag_pattern = '(--?\w+)=?([^\s]+)?'
      flag_indexes = list()
      for i, a in enumerate(args):
         if match_obj := match(flag_pattern, a):
            flag_name, flag_value = match_obj.groups()
            flags[flag_name] = flag_value
            flag_indexes.append(i)

      # Retirando Flags dos Args
      for redutor, i in enumerate(flag_indexes):
         args.pop(i - redutor)
      
      # Executando Comando
      command_function(args, flags)

   except Exception as exc:
      # Exibindo Mensagem de Erro com o Contexto da Exceção
      name = exc.__class__.__name__
      message = exc.args[0]
      messenger.error(message, name)

def start_interactive_mode():
   # Informando que o Modo Interativo foi Iniciado
   global in_interactive_mode
   in_interactive_mode = True

   # Exibindo Mensagem de Boas-Vindas
   show_version()
   messenger.show(message_welcome)

   # Iniciando Loop
   while True:
      # Lendo Argumentos
      args = input('>> ').split()

      # Executando Comando
      if len(args) == 0:
         continue
      command_name = args[0]
      if command_name == 'exit':
         break
      execute_command(command_name, args[1:])

def start(args: list[str]):
   # Iniciando Modo Interativo (Se não houver Argumentos)
   if len(args) == 0:
      start_interactive_mode()

   # Executando Comando Único
   else:
      execute_command(args[0], args[1:])



class CommandError(Exception):
   def __init__(self, message: str, help: bool = False):
      # Verificando se é Necessário Informar o Comando de Ajuda
      if help:
         message += '\nPlease, use help command (lmcv_tools help).'
      super().__init__(message)


import re
from ..interface import searcher

class ResultTable:
   def __init__(self, header: list[str], data: list[list] = []):
      self._header = header.copy()
      self._data = data.copy()
      self.join_fail = False

   # Métodos para Acesso Seguro
   @property
   def header(self):
      return self._header.copy()

   def header_index(self, column_name: str) -> int:
      try:
         return self._header.index(column_name)
      except ValueError:
         raise ValueError(f'The Column name "{column_name}" does not belong in the Result Table Header.')
   
   def verify_row_index(self, index: int):
      if index > (len(self._data) - 1):
         raise IndexError(f'The Row index {index} is out of the Result Table range.')

   # Métodos de Linhas e Colunas
   def column_count(self) -> int:
      return len(self._header)

   def row_count(self) -> int:
      return len(self._data)

   def column(self, name: str) -> list:
      index = self.header_index(name)
      column_data = [row[index] for row in self._data]
      return column_data
      
   def row(self, index: int) -> dict:
      self.verify_row_index(index)
      row_data = {key: value for key, value in zip(self._header, self._data[index])}
      return row_data
   
   def cell(self, column_name: str, row_index) -> str:
      row = self.row(row_index)
      cell = row[column_name]
      return cell
   
   def columns(self):
      for name in self._header:
         yield self.column(name)

   def rows(self):
      for index in range(self.row_count()):
         yield self.row(index)

   def add_row(self, row_data: dict):
      try:
         data_list = [row_data[column_name] for column_name in self._header]
      except KeyError:
         raise KeyError('The Result Table Row Data must contain all header keys.')
      self._data.append(data_list)

   def add_column(self, name: str, default_data):
      self._header.append(name)
      for row in self._data:
         row.append(default_data)

   def delete_row(self, index: int):
      self.verify_row_index(index)
      del self._data[index]
   
   def delete_column(self, name: str):
      index = self.header_index(name)
      del self._header[index]
      for row in self._data:
         del row[index]

   def filter_rows(self, values: dict) -> list[dict]:
      result_rows = self.rows()
      for column, value in values.items():
         result_rows = list(filter(lambda row: row[column] == value, result_rows))
      return result_rows

   # Métodos de Interação com Tabela
   def reorder(self, names: list[str]):
      try:
         data = [[row[name] for name in names] for row in self.rows()]
      except KeyError:
         raise ValueError('Only column names belonging to the Result Table Header must be passed.')
      self._header = names.copy()
      self._data = data.copy()

   def join(self, table):
      # Initializando Variável de Controle para a Falha do Método
      self.join_fail = False

      # Tradando Caso Simples de Tabela Vazia
      if self.column_count() == 0:
         return table

      # Avaliando Atributos Compartilhados pelas Tabelas
      table_1, table_2 = self, table
      attributes_1 = set(table_1._header)
      attributes_2 = set(table_2._header)
      shared_attributes = list(attributes_1 & attributes_2)
      
      # Conservando Tabela Atual Caso não Haja Atributos Compartilhados
      if len(shared_attributes) == 0:
         self.join_fail = True
         return table_1

      # Avaliando se uma das Tabelas já Contém a Outra
      if len(self._header) == len(shared_attributes):
         return table
      if len(table._header) == len(shared_attributes):
         return self
      
      # Instanciando Nova Tabela com todos os Atributos
      joined_columns = list(attributes_1 | attributes_2)
      joined_table = ResultTable(joined_columns)
      
      # Juntando Linhas
      for row_1 in table_1.rows():
         values = {a: row_1[a] for a in shared_attributes}
         rows_2 = table_2.filter_rows(values)
         for row_2 in rows_2:
            row_1.update(row_2)
            joined_table.add_row(row_1)

      return joined_table

   # Métodos de Conversão de Dados da Tabela
   def to_csv(self):
      csv_data = ','.join(self._header)
      for row in self.rows():
         csv_data += '\n' + ','.join(row.values())
      return csv_data

class Operator:
   # Métodos de Avaliação dos Operadores
   def in_evaluate(value, test: str, type_class):
      params = test.split(':')
      len_params = len(params)

      # Caso Onde o Formato dos Parâmetros do Valor de Teste é Inválido
      if len_params > 3 or len_params < 2:
         raise ValueError('The right operand of the "in" operator has a invalid format (it must be [start]:[stop] or [start]:[stop]:[step], all of these integers).')
      
      # Verificando se Todos os Parâmetros são Inteiros
      try:
         params = list(map(int, params))
      except ValueError:
         raise ValueError('All numbers in the right operand of the "in" operator must be integers.')
      
      return type_class(value) in range(*params)

   supported_operators = {
      'and': {
         'type': 'logical',
         'evaluate': lambda c1, c2, av, at: c1.approve(av, at) and c2.approve(av, at)
      },
      'or': {
         'type': 'logical',
         'evaluate': lambda c1, c2, av, at: c1.approve(av, at) or c2.approve(av, at)
      },
      '=': {
         'type': 'relational',
         'evaluate': lambda v, t, tc: tc(v) == tc(t)
      },
      '>': {
         'type': 'relational',
         'evaluate': lambda v, t, tc: tc(v) > tc(t)
      },
      '<': {
         'type': 'relational',
         'evaluate': lambda v, t, tc: tc(v) < tc(t)
      },
      '<=': {
         'type': 'relational',
         'evaluate': lambda v, t, tc: tc(v) <= tc(t)
      },
      '>=': {
         'type': 'relational',
         'evaluate': lambda v, t, tc: tc(v) >= tc(t)
      },
      '!=': {
         'type': 'relational',
         'evaluate': lambda v, t, tc: tc(v) != tc(t)
      },
      'in': {
         'type': 'relational',
         'evaluate': in_evaluate
      }
   }

   def __init__(self, symbol: str):
      try:
         operator = Operator.supported_operators[symbol]
         self.type = operator['type']
         self.evaluate = operator['evaluate']
      except KeyError:
         raise KeyError(f'Unsupported operator "{symbol}" was used.')
   
class Condition:
   def __init__(self, condition_expression: str = None, conditioned_attributes: set = set()):
      # Verificando se a Condição é Nula
      self.is_none_condition = condition_expression is None
      if not self.is_none_condition:
         # Inicializando Conjunto Compartilhado de Atributos Condicionados
         self.conditioned_attributes = conditioned_attributes

         # Buscando Operador na Expressão
         for supported_operator in Operator.supported_operators.keys():
            operands = condition_expression.split(f' {supported_operator} ', 1)
            if len(operands) > 1:
               break
         else:
            raise KeyError('A Condition has not a supported operator.')
         
         # Instanciando Operador e Operandos
         self.operator = Operator(supported_operator)
         if self.operator.type == 'logical':
            self.operand_1 = Condition(operands[0], self.conditioned_attributes)
            self.operand_2 = Condition(operands[1], self.conditioned_attributes)
         else:
            self.conditioned_attributes.add(operands[0])
            self.operand_1 = operands[0]
            self.operand_2 = operands[1]
      
   def approve(self, attribute_values: dict, attribute_types: dict) -> bool:
      # Verificando se a Condição é Nula
      if not self.is_none_condition:
         # Avaliando Condição Lógica
         if self.operator.type == 'logical':
            # Verificano se há Atributos Condicionáveis nos Valores Passados
            set_attributes = set(attribute_values.keys())
            set_conditioned = set(self.conditioned_attributes)
            set_conditionable = set_attributes & set_conditioned

            # Avaliar com Operador se Há Atributos Condicionáveis
            if len(set_conditionable) > 0:
               return self.operator.evaluate(self.operand_1, self.operand_2, attribute_values, attribute_types)
         
         # Avaliando Condição Relacional
         else:
            if self.operand_1 in attribute_values.keys():
               type_class = attribute_types[self.operand_1]
               value = attribute_values[self.operand_1]
               test = self.operand_2
               return self.operator.evaluate(value, test, type_class)
      
      # Retornando True por Padrão
      return True

class Attribute:
   type_relations = {
      'str': str,
      'int': int,
      'float': float,
      'list': list
   }
   extraction_attributes = searcher.get_database('extraction_attributes')

   def __init__(self, name: str):
      # Extraindo Informações do Atributo da Database
      try:
         # Extraindo Index Informado para o Atributo se Houver
         self.name = name
         attribute_index = None
         if result := re.match('([\w.]+)\[(-?\d+)\]', name):
            name = result.group(1)
            attribute_index = int(result.group(2))

         # Atribuindo Informações ao Atributo com base na Database
         info = Attribute.extraction_attributes[name]
         self.format = info['format']

         # Verificando se atributos do tipo "list" tiveram um index informado
         self.type = Attribute.type_relations[info['data_type']]
         if self.type is list:
            if attribute_index is not None:
               self.index = attribute_index
               self.sub_type = Attribute.type_relations[info['sub_type']]
            else:
               raise SyntaxError(f'The Attribute "{name}" must have an associated index.')
            
         # Organizando Atributos Relacionados
         self.related_attributes = dict()
         for related_attribute in info['related_attributes']:
            data_type =  Attribute.extraction_attributes[related_attribute]['data_type']
            type_class = Attribute.type_relations[data_type]
            self.related_attributes[related_attribute] = type_class

      except KeyError:
         raise KeyError(f'Unsupported attribute "{name}" was used.')

   def extract_from(self, pos_data: str, condition: Condition):
      # Instanciando Tabela de Resultado
      result_table = ResultTable([self.name, *self.related_attributes])

      # Extraindo Dados a partir das Keywords
      keyword_matches = re.finditer(self.format['keywords'], pos_data)
      for keyword_match in keyword_matches:
         # Identificando Atributos Relacionados ao Atual a Nível de Keyword
         keyword_related = keyword_match.groupdict()
         line_data = keyword_related.pop('line')
         line_matches = re.finditer(self.format['line'], line_data, re.MULTILINE)

         # Extraindo Dados a partir das Linhas de Keyword
         for line_match in line_matches:
            # Identificando Atributos Relacionados ao Atual a Nível de Linha
            line_related = line_match.groupdict()
            value = line_related.pop('value')

            # Tratamento Especial para Atributos do Tipo "list"
            if self.type is list:
               try:
                  value = value.split()[self.index]
               except IndexError:
                  raise IndexError(f'The Index "[{self.index}]" is out of range in "{self.name}".')

            # Combinando Atributos Relacionados
            line_related.update(keyword_related)

            # Criando Dicionários com Todos os Valores do Atributo (Próprio e Relacionados)
            attribute_values = dict()
            attribute_values[self.name] = value
            for key, value in line_related.items():
               attribute_values[key.replace('_', '.')] = value

            # Criando Dicionários com Todos os Tipos do Atributo (Próprio e Relacionados)
            attribute_types = dict()
            attribute_types[self.name] = self.type if (self.type is not list) else self.sub_type
            attribute_types.update(self.related_attributes)

            # Aprovando ou Não Valores para a Tabela de Resultado com Base na Condição
            if condition.approve(attribute_values, attribute_types):
               result_table.add_row(attribute_values)
      
      return result_table

   def change_to(self, value_to_change: str, dat_data: str, condition: Condition):
      # Verificando se Atributos Condicionados são Relacionados com o Atributo a ser Alterado
      related_attributes = set(self.related_attributes.keys())
      related_attributes.add(self.name)
      if (
         (not condition.is_none_condition) and 
         (len(condition.conditioned_attributes | related_attributes) > len(related_attributes))
      ):
         raise KeyError('All Conditioned Attributes must be related to the Attribute that will be changed.') 

      # Percorrendo Keywords do Atributo
      keyword_matches = re.finditer(self.format['keywords'], dat_data)
      for keyword_match in keyword_matches:
         # Identificando Atributos Relacionados ao Atual a Nível de Keyword
         keyword_related = keyword_match.groupdict()
         line_data = keyword_related.pop('line')
         line_matches = re.finditer(self.format['line'], line_data, re.MULTILINE)

         # Extraindo Dados a partir das Linhas de Keyword
         new_line_data = line_data
         for line_match in line_matches:
            # Identificando Atributos Relacionados ao Atual a Nível de Linha
            line_related = line_match.groupdict()
            value = line_related.pop('value')

            # Tratamento Especial para Atributos do Tipo "list"
            if self.type is list:
               try:
                  value = value.split()[self.index]
               except IndexError:
                  raise IndexError(f'The Index "[{self.index}]" is out of range in "{self.name}".')

            # Combinando Atributos Relacionados
            line_related.update(keyword_related)

            # Criando Dicionários com Todos os Valores do Atributo (Próprio e Relacionados)
            attribute_values = dict()
            attribute_values[self.name] = value
            for key, value in line_related.items():
               attribute_values[key.replace('_', '.')] = value

            # Criando Dicionários com Todos os Tipos do Atributo (Próprio e Relacionados)
            attribute_types = dict()
            attribute_types[self.name] = self.type if (self.type is not list) else self.sub_type
            attribute_types.update(self.related_attributes)

            # Aprovando ou Não Valores para a Alteração com Base na Condição
            if condition.approve(attribute_values, attribute_types):
               # Instanciando Linha Antiga e Linha Nova
               old_line = line_match.group()
               new_line = self.format['line']

               # Substituindo Formatos Regex pelos Valores para a Nova Linha
               attribute_values.pop(self.name)
               attribute_values['value'] = value_to_change
               for name, value in attribute_values.items():
                  name = name.replace('.', '_')
                  new_line = re.sub(f'\(\?P<{name}>[^\)]+[^\(]+\)', value, new_line)
               new_line = new_line.replace('\s+', '   ')
               
               # Substituindo Linha Antiga pela Nova no Conjunto de Linhas
               new_line_data = new_line_data.replace(old_line, new_line)

         # Substituindo Conjunto de Linhas Antigo pelo Novo no .dat
         dat_data = dat_data.replace(line_data, new_line_data)

      return dat_data



from math import (
    pi,
    sin,
    cos,
    factorial,
    comb
)
from ..models.math import (
    Matrix
)

# --------------------------------------------------
# 1 - B-Splines - Classes Relacionadas
# --------------------------------------------------
class BSpline_Basis:
    def __init__(self, degree: int, knot_vector: list[float]) -> None:
        self.degree = degree
        self.knot_vector = knot_vector.copy()
        self.n_basis = len(knot_vector) - degree - 1
    
    def alpha(self, k: int, knot: float, i: int):
        p = self.degree
        if i <= (k - p):
            alpha = 1
        elif (i >= (k - p + 1)) and (i <= k):
            alpha = knot - self.knot_vector[i]
            alpha /= (self.knot_vector[i + p] - self.knot_vector[i])
        elif i >= (k + 1):
            alpha = 0
        return alpha

    def __call__(self, p: int, i: int, t: float):
        # Recursão - Caso Base
        if p == 0:
            if t >= self.knot_vector[i] and t <= self.knot_vector[i + 1]:
                return 1
            else:
                return 0
            
        # Recursão - Caso Geral
        N1 = t - self.knot_vector[i]
        D1 = self.knot_vector[i + p] - self.knot_vector[i]
        N2 = self.knot_vector[i + p + 1] - t
        D2 = self.knot_vector[i + p + 1] - self.knot_vector[i + 1]
        N = 0.0
        if D1 != 0.0:
            N += N1 / D1 * self.__call__(p - 1, i, t)
        if D2 != 0.0:
            N += N2 / D2 * self.__call__(p - 1, i + 1, t)
        return N

class BSpline_Curve:
    def __init__(self, degree: int, knot_vector: list[float], control_points: list[list]) -> None:
        self.basis = BSpline_Basis(degree, knot_vector)
        self.control_points = control_points.copy()
        self.dimension = len(control_points[0])
        
    def knot_insertion(self, k: int, knot: float):
        # Adicionando Knot
        new_knot_vector = self.basis.knot_vector.copy()
        new_knot_vector.insert(k + 1, knot)

        # Criando Novos Pontos de Controle
        new_control_points = list()
        for i in range(0, self.basis.n_basis + 1):
            # Determinando Alfa
            alpha = self.basis.alpha(k, knot, i)

            # Calculando Pontos
            control_point = [0] * self.dimension
            for d in range(self.dimension):
                if alpha:
                    control_point[d] += alpha * self.control_points[i][d]
                if (1 - alpha):
                    control_point[d] += (1 - alpha) * self.control_points[i - 1][d]
            new_control_points.append(control_point.copy())
        
        # Atualizando Atributos
        self.basis.n_basis += 1
        self.basis.knot_vector = new_knot_vector
        self.control_points = new_control_points
    
    def knot_insertions(self, knots: list[float]):
        for knot in reversed(knots):
            self.knot_insertion(self.basis.degree, knot)
    
    def degree_elevation(self, times: float):
        # Verificando se a B-Spline é Equivalente à uma Curva de Bézier
        if len(set(self.basis.knot_vector)) != 2:
            raise RuntimeError('The Curve must be a Bézier Equivalent Curve.')

        # Calculando Novo Número de Funções de Base
        t = times
        s = len(set(self.basis.knot_vector)) - 2
        new_n_basis = self.basis.n_basis + t * (s + 1)

        # Calculando Novo Vetor de Knots
        new_knot_vector = list()
        added_knots = set()
        for knot in self.basis.knot_vector:
            new_multiplicity = self.basis.knot_vector.count(knot) + t
            if knot not in added_knots:
                new_knot_vector.extend([knot] * new_multiplicity)
                added_knots.add(knot)

        # Criando Novos Pontos de Controle
        p = self.basis.degree
        new_control_points = list()
        for i in range(0, p + t + 1):
            control_point = [0] * self.dimension
            for d in range(self.dimension):
                control_point[d] = sum(
                    comb(p, j) * comb(t, i - j) * self.control_points[j][d] / comb(p + t, i)
                    for j in range(
                        max(0, i - t),
                        min(p, i) + 1
                    )
                )
            new_control_points.append(control_point)
        
        # Atualizando Atributos
        self.basis.degree += t
        self.basis.n_basis = new_n_basis
        self.basis.knot_vector = new_knot_vector
        self.control_points = new_control_points
        
    def __call__(self, t: float):
        point = list()
        for d in range(self.dimension):
            point.append(sum([
                self.basis(self.basis.degree, i, t) * self.control_points[i][d]
                for i in range(self.basis.n_basis)
            ]))
        return point
    
class BSpline_Surface:
    def __init__(
        self, 
        degree: list[int],
        knot_vectors: list[list[float]],
        control_points: list[list[list]]
    ) -> None:
        self.basis = [BSpline_Basis(p, kv) for p, kv in zip(degree, knot_vectors)]
        self.control_points = control_points.copy()
        self.dimension = len(control_points[0][0])
        
    def __call__(self, u: float, v: float):
        point = list()
        for d in range(self.dimension):
            point.append(sum([
                (
                    self.basis[0](self.degree[0], i, u) *
                    self.basis[1](self.degree[1], j, v) *
                    self.control_points[i][j][d]
                )
                for j in range(self.n_basis[1])
                for i in range(self.n_basis[0])
            ]))
        return point

# --------------------------------------------------
# 2 - NURBS - Classes Relacionadas
# --------------------------------------------------
class NURBS_Curve(BSpline_Curve):
    def __init__(
        self, 
        degree: int, 
        knot_vector: list[float], 
        control_points: list[list],
        weights: list[float]
    ) -> None:
        # Chamando Construtor da Super-Classe
        super().__init__(degree, knot_vector, control_points)

        # Conferindo Pesos dos Pontos de Controle
        if len(weights) != len(control_points):
            raise ValueError('The number of weights and the number of control points must be the same.')
        self.weights = weights.copy()
    
    def knot_insertion(self, k: int, knot: float):
        # Fazer Inserção de Knot para Pesos
        new_weights = list()
        for i in range(0, self.basis.n_basis + 1):
            # Determinando Alfa
            alpha = self.basis.alpha(k, knot, i)

            # Calculando Peso
            weight = 0.0
            if alpha:
                weight += alpha * self.weights[i]
            if (1 - alpha):
                weight += (1 - alpha) * self.weights[i - 1]
            new_weights.append(weight)

        # Atualizando Pontos de Controle a Partir dos Pesos
        for d in range(self.dimension):
            for i in range(len(self.control_points)):
                self.control_points[i][d] *= self.weights[i]

        # Fazendo Inserção de Knot da Super-Classe
        super().knot_insertion(k, knot)

        # Atualizando Pesos
        self.weights = new_weights.copy()

        # Atualizando Pontos de Controle a Partir dos Pesos
        for d in range(self.dimension):
            for i in range(len(self.control_points)):
                self.control_points[i][d] /= self.weights[i]
    
    def degree_elevation(self, times: float):
        # Fazendo Elevação de Grau para Pesos
        t = times
        p = self.basis.degree
        s = len(set(self.basis.knot_vector)) - 2
        new_weights = list()
        for i in range(0, p + s + t + 1):
            weight = sum(
                comb(p, j) * comb(t, i - j) * self.weights[j] / comb(p + t, i)
                for j in range(
                    max(0, i - t),
                    min(p, i) + 1
                )
            )
            new_weights.append(weight)
        
        # Atualizando Pontos de Controle a Partir dos Pesos
        for d in range(self.dimension):
            for i in range(len(self.control_points)):
                self.control_points[i][d] *= self.weights[i]

        # Fazendo Elevação de Grau da Super-Classe
        super().degree_elevation(times)

        # Atualizando Pesos
        self.weights = new_weights.copy()

        # Atualizando Pontos de Controle a Partir dos Pesos
        for d in range(self.dimension):
            for i in range(len(self.control_points)):
                self.control_points[i][d] /= self.weights[i]

    def __call__(self, t: float):
        # Inicializando Ponto
        point = list()

        # Calculando Denominador Comum da NURBS
        D = sum([
            self.basis(self.degree, i, t) * self.weights[i]
            for i in range(self.basis.n_basis)
        ])

        # Calculando Cada Dimensão do Ponto
        for d in range(self.dimension):
            N = sum([
                self.basis(self.basis.degree, i, t) * self.control_points[i][d] * self.weights[i]
                for i in range(self.basis.n_basis)
            ])
            point.append(N / D)
        
        return point

class NURBS_Surface(BSpline_Surface):
    def __init__(
        self, 
        degree: list[int],
        knot_vectors: list[list[float]],
        control_points: list[list[list]],
        weights: list[list[float]]
    ) -> None:
        super().__init__(degree, knot_vectors, control_points)
        self.weights = weights
        
    def __call__(self, u: float, v: float) -> list[float]:
        point = list()

        # Calculando Denominador Comum da NURBS
        D = sum([
            self.basis[0](self.basis[0].degree, i, u) *
            self.basis[1](self.basis[1].degree, j, v) *
            self.weights[i][j]
            for j in range(self.basis[1].n_basis)
            for i in range(self.basis[0].n_basis)
        ])

        for d in range(self.dimension):
            N = sum([
                (
                    self.basis[0](self.basis[0].degree, i, u) *
                    self.basis[1](self.basis[1].degree, j, v) *
                    self.control_points[i][j][d] *
                    self.weights[i][j]
                )
                for j in range(self.basis[1].n_basis)
                for i in range(self.basis[0].n_basis)
            ])
            point.append(N / D)
        return point

# --------------------------------------------------
# 3 - Curvas de Bézier - Funções Relacionadas
# --------------------------------------------------
def bezier_equiv_coord(c: float, c0: float, c2: float):
   return 2 * c - 0.5 * (c0 + c2)

def bernstein_polynomial(index: int, grade: int, region: float):
   # Renomeando Parâmetros para Facilitar os Cálculos
   i, p, t = index, grade, region
   
   # Verificando Validade dos Parâmetros
   if i < 0 or i > p:
      raise ValueError(f'Index {i} does not exist for Bernstein Polynomial with Grade {p}.')
   
   # Calculando Polinômio na Região Informada
   return (factorial(p) / (factorial(i) * factorial(p - i))) * t ** i * (1 - t) ** (p - i)

# --------------------------------------------------
# 4 - Transformador - Classes Relacionadas
# --------------------------------------------------
class GeometricalTransformer:
    # Matrizes de Rotação
    def Rx(self, angle: float):
        angle = angle * pi / 180
        Rot = Matrix(4, 4)
        Rot[0, 0] = 1
        Rot[1, 1] = cos(angle)
        Rot[1, 2] = sin(angle)
        Rot[2, 1] = -sin(angle)
        Rot[2, 2] = cos(angle)
        Rot[3, 3] = 1
        return Rot
    
    def Ry(self, angle: float):
        angle = angle * pi / 180
        Rot = Matrix(4, 4)
        Rot[0, 0] = cos(angle)
        Rot[0, 2] = -sin(angle)
        Rot[1, 1] = 1
        Rot[2, 0] = sin(angle)
        Rot[2, 2] = cos(angle)
        Rot[3, 3] = 1
        return Rot
    
    def Rz(self, angle: float):
        angle = angle * pi / 180
        Rot = Matrix(4, 4)
        Rot[0, 0] = cos(angle)
        Rot[0, 1] = sin(angle)
        Rot[1, 0] = -sin(angle)
        Rot[1, 1] = cos(angle)
        Rot[2, 2] = 1
        Rot[3, 3] = 1
        return Rot
    
    # Métodos Públicos
    def calculate_centroid(self, points: list[list[float]]) -> list[float]:
        n_points = len(points)
        n_coordinates = len(points[0])
        centroid = n_coordinates * [0]
        for point in points:
            for i in range(n_coordinates):
                centroid[i] += point[i]
        for i in range(n_coordinates):
            centroid[i] /= n_points
        return centroid
    
    def calculate_colinearity(self, points: list[list[float]]) -> float:
      factor = 0
      for i in range(0, len(points) - 2):
         diag1 = points[i][0] * points[i + 1][1] + points[i + 1][0] * points[i + 2][1] + points[i + 2][0]  * points[i][1]
         diag2 = points[i][0] * points[i + 2][1] + points[i + 1][0] * points[i][1] + points[i + 2][0]  * points[i + 1][1]
         factor += abs(diag1 - diag2)
      return abs(factor)

    def homogenize_coordinates(self, x: float, y: float, z: float, h: float) -> Matrix:
        h_matrix = Matrix(1, 4)
        h_matrix[0, 0] = x * h
        h_matrix[0, 1] = y * h
        h_matrix[0, 2] = z * h
        h_matrix[0, 3] = 1
        return h_matrix
    
    def heterogenize_coordinates(self, h_matrix: Matrix) -> list[float]:
        if h_matrix.n_rows != 1 or h_matrix.n_colmuns != 4:
            raise ValueError('The homogenized coordinates matrix must have 1 row and 4 columns.')
        h_matrix = h_matrix.to_list()[0]
        coordinates = [h_matrix[i] / h_matrix[3] for i in range(3)]
        return coordinates
    
    def translate(self, x: float, y: float, z: float, tx: float, ty: float, tz: float):
        # Homogenizando Coordenadas
        h = self.homogenize_coordinates(x, y, z, 1)

        # Gerando Matriz de Translação 3D
        Trans = Matrix(4, 4)
        Trans.fill_diagonal(1)
        Trans[3, 0] = tx
        Trans[3, 1] = ty
        Trans[3, 2] = tz

        # Transladando e Retornando Pontos
        ht = h * Trans
        return self.heterogenize_coordinates(ht)

    def rotate(self, x: float, y: float, z: float, rotations: list[Matrix]) -> tuple[float]:
        # Homogenizando Coordenadas
        h = self.homogenize_coordinates(x, y, z, 1)

        # Gerando Matriz de Rotação 3D
        Rot = Matrix(4, 4)
        Rot.fill_diagonal(1.0)
        for rotation in rotations:
            Rot = Rot * rotation

        # Rotacionando e Retornando Pontos
        hr = h * Rot
        return self.heterogenize_coordinates(hr)
    
    def project_parallel(self, x: float, y: float, z: float) -> tuple[float]:
        return x, y
    
    def project_perspective(self, x: float, y: float, z: float, x_cop: float, y_cop: float, z_cop: float) -> tuple[float]:
        # Homogenizando Coordenadas
        h = self.homogenize_coordinates(x, y, z, 1)

        # Gerando Matriz de Projeção
        P = Matrix(4, 4)
        P[0, 0] = 1
        P[1, 1] = 1
        P[2, 0] = - x_cop / z_cop 
        P[2, 1] = - y_cop / z_cop 
        P[2, 3] = - 1 / z_cop 
        P[3, 3] = 1

        # Projetando e Retornando Pontos
        hp = h * P
        u, v, n = self.heterogenize_coordinates(hp)
        return u, v


from inspect import signature
from ..interface import filer
from ..models.graphics import PromptGenerateVirtualLaminas
from ..models.custom_errors import CommandError
from ..models.artifacts import (
   VirtualLaminas,
   ElementConfiguration,
   Rectangle,
   Cuboid,
   NURBS_Rectangle,
   NURBS_Cuboid,
   CylindricalPanel,
   SlitAnnularPlate
)
from ..models.simulation import (
   IsotropicMaterial,
   FunctionallyGradedMaterial
)

# Funções de Geração de Artefatos
def generate_virtual_laminas(
   laminas_count: int,
   element_type: str,
   thickness: float,
   number_integration_points: int,
   power_law_exponent: float,
   micromechanical_model: str,
   E1: float,
   E2: float,
   nu1: float,
   nu2: float,
   rho1: float,
   rho2: float,
   smart: bool
) -> VirtualLaminas:
   # Instanciando Configuração do Elemento
   element = ElementConfiguration(element_type, number_integration_points)

   # Instanciando Materiais
   materials = list()
   materials.append(IsotropicMaterial(E1, nu1, rho1))
   materials.append(IsotropicMaterial(E2, nu2, rho2))

   # Instanciando FGM
   fgm = FunctionallyGradedMaterial(micromechanical_model, materials)
   
   # Instanciando Artefato de Lâminas Virtuais
   virtual_laminas = VirtualLaminas(
      laminas_count,
      thickness,
      power_law_exponent,
      element,
      fgm,
      smart
   )
   virtual_laminas.generate()

   return virtual_laminas

def generate_rectangle(
   element_type: str, 
   dimensions: list[float], 
   discretization: list[int]
) -> Rectangle:
   # Instanciando Artefato
   rectangle = Rectangle(element_type, dimensions, discretization)

   # Gerando Artefato
   rectangle.generate()

   # Retornando Artefato
   return rectangle

def generate_cuboid(
   element_type: str, 
   dimensions: list[float], 
   discretization: list[int]
) -> Cuboid:
   # Instanciando Artefato
   cuboid = Cuboid(element_type, dimensions, discretization)

   # Gerando Artefato
   cuboid.generate()

   # Retornando Artefato
   return cuboid

def generate_nurbs_rectangle(
   degrees: list[int],
   dimensions: list[float],
   discretization: list[int]
) -> NURBS_Rectangle:
   # Instanciando Artefato
   nurbs_rectangle = NURBS_Rectangle(degrees, dimensions, discretization)

   # Gerando Artefato
   nurbs_rectangle.generate()

   # Retornando Artefato
   return nurbs_rectangle

def generate_nurbs_cuboid(
   degrees: list[int],
   dimensions: list[float],
   discretization: list[int]
) -> NURBS_Cuboid:
   # Instanciando Artefato
   nurbs_cuboid = NURBS_Cuboid(degrees, dimensions, discretization)

   # Gerando Artefato
   nurbs_cuboid.generate()

   # Retornando Artefato
   return nurbs_cuboid

def generate_cyl_panel(
   element_type: str,
   height: float,
   discretization: list[int],
   radius: float,
   angles: list[float]
) -> CylindricalPanel:
   # Instanciando Artefato
   cyl_panel = CylindricalPanel(element_type, height, discretization, radius, angles)

   # Gerando Artefato
   cyl_panel.generate()

   # Retornando Artefato
   return cyl_panel

def generate_slit_annular_plate(
   element_type: str,
   inner_radius: float,
   outer_radius: float,
   discretization: list[int]
) -> SlitAnnularPlate:
   # Instanciando Artefato
   slit_annular_plate = SlitAnnularPlate(element_type, inner_radius, outer_radius, discretization)

   # Gerando Artefato
   slit_annular_plate.generate()

   # Retornando Artefato
   return slit_annular_plate

# Funções de Parâmetros de Artefatos
def params_virtual_laminas(args: list[str]) -> dict:
   # Iniciando Parâmetros
   params = dict()
   
   # Exibindo Interface Gráfica para Preencher Parâmetros (Se não Houver Parâmetros)
   if len(args) == 0:
      window = PromptGenerateVirtualLaminas(params)
      window.start()

      # Conferindo se Há um Caminho
      if params.get('path') is not None:
         if params['path'] != '':
            args = [params['path']]
         del params['path']

   # Tentando Coletar Parâmetros Dados 
   else:
      try:
         reference = dict(signature(generate_virtual_laminas).parameters)
         index = 0
         for name, param_obj in reference.items():
            type_class = param_obj.annotation
            if type_class is bool:
               params[name] = False if args.pop(0) == 'False' else True
            else:   
               params[name] = type_class(args.pop(0))
            index += 1
      except IndexError:
         raise CommandError('Invalid number of arguments.', help=True)
   
   return params, args

def params_rectangle(args: list[str]) -> dict:
   # Iniciando Parâmetros
   params = dict()

   # Tentando Converter Tipos de Dados
   try:
      params['element_type'] = args[0]
   except IndexError:
      raise ValueError(f'A Rectangle needs an element type.')
   params['dimensions'] = list(map(float, args[1:3]))
   params['discretization'] = list(map(int, args[3:5]))
   
   return params, args[5:]

def params_cuboid(args: list[str]) -> dict:
   # Iniciando Parâmetros
   params = dict()

   # Tentando Converter Tipos de Dados
   try:
      params['element_type'] = args[0]
   except IndexError:
      raise ValueError(f'A Cuboid needs an element type.')
   params['dimensions'] = list(map(float, args[1:4]))
   params['discretization'] = list(map(int, args[4:7]))
   
   return params, args[7:]

def params_nurbs_rectangle(args: list[str]) -> dict:
   # Iniciando Parâmetros
   params = dict()

   # Tentando Converter Tipos de Dados
   params['degrees'] = list(map(int, args[0:2]))
   params['dimensions'] = list(map(float, args[2:4]))
   params['discretization'] = list(map(int, args[4:6]))
   
   return params, args[6:]

def params_nurbs_cuboid(args: list[str]) -> dict:
   # Iniciando Parâmetros
   params = dict()

   # Tentando Converter Tipos de Dados
   params['degrees'] = list(map(int, args[0:3]))
   params['dimensions'] = list(map(float, args[3:6]))
   params['discretization'] = list(map(int, args[6:9]))
   
   return params, args[9:]

def params_cyl_panel(args: list[str]) -> dict:
   # Iniciando Parâmetros
   params = dict()

   # Tentando Converter Tipos de Dados
   try:
      params['element_type'] = args[0]
   except IndexError:
      raise ValueError(f'A Cylindrical Panel needs an element type.')
   try:
      params['height'] = float(args[1])
   except IndexError:
      raise ValueError(f'A Cylindrical Panel needs a height.')
   params['discretization'] = list(map(int, args[2:4]))
   try:
      params['radius'] = float(args[4])
   except IndexError:
      raise ValueError(f'A Cylindrical Panel needs a radius.')
   params['angles'] = list(map(float, args[5:7]))
   
   return params, args[7:]

def params_slit_annular_plate(args: list[str]) -> dict:
   # Iniciando Parâmetros
   params = dict()

   # Tentando Converter Tipos de Dados
   try:
      params['element_type'] = args[0]
   except IndexError:
      raise ValueError(f'A Slit Annular Plate needs an element type.')
   try:
      params['inner_radius'] = float(args[1])
   except IndexError:
      raise ValueError(f'A Slit Annular Plate Panel needs a inner radius.')
   try:
      params['outer_radius'] = float(args[2])
   except IndexError:
      raise ValueError(f'A Slit Annular Plate Panel needs a outer radius.')
   params['discretization'] = list(map(int, args[3:5]))
   
   return params, args[5:]

# Relação Artefato/Funções
artifacts = {
   'virtual_laminas': {
      'params': params_virtual_laminas,
      'generate': generate_virtual_laminas
   },
   'rectangle': {
      'params': params_rectangle,
      'generate': generate_rectangle
   },
   'cuboid': {
      'params': params_cuboid,
      'generate': generate_cuboid
   },
   'nurbs_rectangle': {
      'params': params_nurbs_rectangle,
      'generate': generate_nurbs_rectangle
   },
   'nurbs_cuboid': {
      'params': params_nurbs_cuboid,
      'generate': generate_nurbs_cuboid
   },
   'cyl_panel': {
      'params': params_cyl_panel,
      'generate': generate_cyl_panel
   },
   'slit_annular_plate': {
      'params': params_slit_annular_plate,
      'generate': generate_slit_annular_plate
   }
}

# Funções de Inicialização
def start(artifact_name: str, args: list[str]) -> str:
   try:
      # Coletando Parâmetros
      params_function = artifacts[artifact_name]['params']
      params, args = params_function(args)

      # Gerando Artefato
      generate_function = artifacts[artifact_name]['generate']
      artifact = generate_function(**params)

   except KeyError:
      raise CommandError(f'Unknown Artifact "{artifact_name}"')
   except TypeError:
      raise CommandError('Not all arguments were correctly passed.')

   # Escrevendo Arquivo do Artefato
   try:
      path = args[0]
   except IndexError:
      path = artifact.file_name
   filer.write(path, artifact.data)



# Importações - Bibliotecas Nativas
import json
from importlib.resources import files

# Constantes Globais
databases_files = files('lmcv_tools.resources.databases')

# Funções Globais
def get_database(json_name: str) -> dict:
   # Abrindo Arquivo JSON da Base de Dados
   json_file = databases_files.joinpath(json_name + '.json')
   json_file = json_file.open('r')
   json_data = json.load(json_file)
   return json_data
